
# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class AccountData(common.Structure):
	def __init__(self):
		super().__init__()
		self.pid = None
		self.name = None
		self.groups = None
		self.email = None
		self.creation_date = None
		self.effective_date = None
		self.not_effective_message = None
		self.expiry_date = None
		self.expired_message = None
	
	def check_required(self, settings, version):
		for field in ['pid', 'name', 'groups', 'email', 'creation_date', 'effective_date', 'not_effective_message', 'expiry_date', 'expired_message']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.pid = stream.pid()
		self.name = stream.string()
		self.groups = stream.u32()
		self.email = stream.string()
		self.creation_date = stream.datetime()
		self.effective_date = stream.datetime()
		self.not_effective_message = stream.string()
		self.expiry_date = stream.datetime()
		self.expired_message = stream.string()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.pid(self.pid)
		stream.string(self.name)
		stream.u32(self.groups)
		stream.string(self.email)
		stream.datetime(self.creation_date)
		stream.datetime(self.effective_date)
		stream.string(self.not_effective_message)
		stream.datetime(self.expiry_date)
		stream.string(self.expired_message)


class BasicAccountInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.pid = None
		self.name = None
	
	def check_required(self, settings, version):
		for field in ['pid', 'name']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.pid = stream.pid()
		self.name = stream.string()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.pid(self.pid)
		stream.string(self.name)


class AccountProtocol:
	METHOD_CREATE_ACCOUNT = 1
	METHOD_DELETE_ACCOUNT = 2
	METHOD_DISABLE_ACCOUNT = 3
	METHOD_CHANGE_PASSWORD = 4
	METHOD_TEST_CAPABILITY = 5
	METHOD_GET_NAME = 6
	METHOD_GET_ACCOUNT_DATA = 7
	METHOD_GET_PRIVATE_DATA = 8
	METHOD_GET_PUBLIC_DATA = 9
	METHOD_GET_MULTIPLE_PUBLIC_DATA = 10
	METHOD_UPDATE_ACCOUNT_NAME = 11
	METHOD_UPDATE_ACCOUNT_EMAIL = 12
	METHOD_UPDATE_CUSTOM_DATA = 13
	METHOD_FIND_BY_NAME_REGEX = 14
	METHOD_UPDATE_ACCOUNT_EXPIRY_DATE = 15
	METHOD_UPDATE_ACCOUNT_EFFECTIVE_DATE = 16
	METHOD_UPDATE_STATUS = 17
	METHOD_GET_STATUS = 18
	METHOD_GET_LAST_CONNECTION_STATS = 19
	METHOD_RESET_PASSWORD = 20
	METHOD_CREATE_ACCOUNT_WITH_CUSTOM_DATA = 21
	METHOD_RETRIEVE_ACCOUNT = 22
	METHOD_UPDATE_ACCOUNT = 23
	METHOD_CHANGE_PASSWORD_BY_GUEST = 24
	METHOD_FIND_BY_NAME_LIKE = 25
	METHOD_CUSTOM_CREATE_ACCOUNT = 26
	METHOD_NINTENDO_CREATE_ACCOUNT = 27
	METHOD_LOOKUP_OR_CREATE_ACCOUNT = 28
	METHOD_DISCONNECT_PRINCIPAL = 29
	METHOD_DISCONNECT_ALL_PRINCIPALS = 30
	
	PROTOCOL_ID = 0x19


class AccountClient(AccountProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def create_account(self, name, key, groups, email):
		logger.info("AccountClient.create_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		stream.string(key)
		stream.u32(groups)
		stream.string(email)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CREATE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.result()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.create_account -> done")
		return result
	
	async def delete_account(self, pid):
		logger.info("AccountClient.delete_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.delete_account -> done")
	
	async def disable_account(self, pid, until, message):
		logger.info("AccountClient.disable_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		stream.datetime(until)
		stream.string(message)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DISABLE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.result()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.disable_account -> done")
		return result
	
	async def change_password(self, new_key):
		logger.info("AccountClient.change_password()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(new_key)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CHANGE_PASSWORD, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.change_password -> done")
		return result
	
	async def test_capability(self, capability):
		logger.info("AccountClient.test_capability()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(capability)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_TEST_CAPABILITY, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.test_capability -> done")
		return result
	
	async def get_name(self, pid):
		logger.info("AccountClient.get_name()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_NAME, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		name = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_name -> done")
		return name
	
	async def get_account_data(self):
		logger.info("AccountClient.get_account_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_ACCOUNT_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.data = stream.extract(AccountData)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_account_data -> done")
		return obj
	
	async def get_private_data(self):
		logger.info("AccountClient.get_private_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_PRIVATE_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.bool()
		obj.data = stream.anydata()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_private_data -> done")
		return obj
	
	async def get_public_data(self, pid):
		logger.info("AccountClient.get_public_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_PUBLIC_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.bool()
		obj.data = stream.anydata()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_public_data -> done")
		return obj
	
	async def get_multiple_public_data(self, pids):
		logger.info("AccountClient.get_multiple_public_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(pids, stream.pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_MULTIPLE_PUBLIC_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.bool()
		obj.data = stream.list(stream.anydata)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_multiple_public_data -> done")
		return obj
	
	async def update_account_name(self, name):
		logger.info("AccountClient.update_account_name()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_ACCOUNT_NAME, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.result()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_account_name -> done")
		return result
	
	async def update_account_email(self, email):
		logger.info("AccountClient.update_account_email()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(email)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_ACCOUNT_EMAIL, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.result()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_account_email -> done")
		return result
	
	async def update_custom_data(self, public_data, private_data):
		logger.info("AccountClient.update_custom_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.anydata(public_data)
		stream.anydata(private_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_CUSTOM_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.result()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_custom_data -> done")
		return result
	
	async def find_by_name_regex(self, groups, regex, range):
		logger.info("AccountClient.find_by_name_regex()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(groups)
		stream.string(regex)
		stream.add(range)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_FIND_BY_NAME_REGEX, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		accounts = stream.list(BasicAccountInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.find_by_name_regex -> done")
		return accounts
	
	async def update_account_expiry_date(self, pid, expiry, message):
		logger.info("AccountClient.update_account_expiry_date()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		stream.datetime(expiry)
		stream.string(message)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_ACCOUNT_EXPIRY_DATE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_account_expiry_date -> done")
	
	async def update_account_effective_date(self, pid, effective_from, message):
		logger.info("AccountClient.update_account_effective_date()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		stream.datetime(effective_from)
		stream.string(message)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_ACCOUNT_EFFECTIVE_DATE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_account_effective_date -> done")
	
	async def update_status(self, status):
		logger.info("AccountClient.update_status()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(status)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_STATUS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_status -> done")
	
	async def get_status(self, pid):
		logger.info("AccountClient.get_status()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_STATUS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		status = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_status -> done")
		return status
	
	async def get_last_connection_stats(self, pid):
		logger.info("AccountClient.get_last_connection_stats()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_LAST_CONNECTION_STATS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.last_session_login = stream.datetime()
		obj.last_session_logout = stream.datetime()
		obj.current_session_login = stream.datetime()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.get_last_connection_stats -> done")
		return obj
	
	async def reset_password(self):
		logger.info("AccountClient.reset_password()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_RESET_PASSWORD, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.reset_password -> done")
		return result
	
	async def create_account_with_custom_data(self, name, key, groups, email, public_data, private_data):
		logger.info("AccountClient.create_account_with_custom_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		stream.string(key)
		stream.u32(groups)
		stream.string(email)
		stream.anydata(public_data)
		stream.anydata(private_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CREATE_ACCOUNT_WITH_CUSTOM_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.create_account_with_custom_data -> done")
	
	async def retrieve_account(self):
		logger.info("AccountClient.retrieve_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_RETRIEVE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.account_data = stream.extract(AccountData)
		obj.public_data = stream.anydata()
		obj.private_data = stream.anydata()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.retrieve_account -> done")
		return obj
	
	async def update_account(self, key, email, public_data, private_data):
		logger.info("AccountClient.update_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(key)
		stream.string(email)
		stream.anydata(public_data)
		stream.anydata(private_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.update_account -> done")
	
	async def change_password_by_guest(self, name, email, key):
		logger.info("AccountClient.change_password_by_guest()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		stream.string(email)
		stream.string(key)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CHANGE_PASSWORD_BY_GUEST, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.change_password_by_guest -> done")
	
	async def find_by_name_like(self, groups, like, range):
		logger.info("AccountClient.find_by_name_like()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(groups)
		stream.string(like)
		stream.add(range)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_FIND_BY_NAME_LIKE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		accounts = stream.list(BasicAccountInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.find_by_name_like -> done")
		return accounts
	
	async def custom_create_account(self, name, key, groups, email, auth_data):
		logger.info("AccountClient.custom_create_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		stream.string(key)
		stream.u32(groups)
		stream.string(email)
		stream.anydata(auth_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CUSTOM_CREATE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		pid = stream.pid()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.custom_create_account -> done")
		return pid
	
	async def nintendo_create_account(self, name, key, groups, email, auth_data):
		logger.info("AccountClient.nintendo_create_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		stream.string(key)
		stream.u32(groups)
		stream.string(email)
		stream.anydata(auth_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_NINTENDO_CREATE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.pid = stream.pid()
		obj.pid_hmac = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.nintendo_create_account -> done")
		return obj
	
	async def lookup_or_create_account(self, name, key, groups, email, auth_data):
		logger.info("AccountClient.lookup_or_create_account()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(name)
		stream.string(key)
		stream.u32(groups)
		stream.string(email)
		stream.anydata(auth_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_LOOKUP_OR_CREATE_ACCOUNT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		pid = stream.pid()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.lookup_or_create_account -> done")
		return pid
	
	async def disconnect_principal(self, pid):
		logger.info("AccountClient.disconnect_principal()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DISCONNECT_PRINCIPAL, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.disconnect_principal -> done")
		return result
	
	async def disconnect_all_principals(self):
		logger.info("AccountClient.disconnect_all_principals()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DISCONNECT_ALL_PRINCIPALS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AccountClient.disconnect_all_principals -> done")
		return result


class AccountServer(AccountProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_CREATE_ACCOUNT: self.handle_create_account,
			self.METHOD_DELETE_ACCOUNT: self.handle_delete_account,
			self.METHOD_DISABLE_ACCOUNT: self.handle_disable_account,
			self.METHOD_CHANGE_PASSWORD: self.handle_change_password,
			self.METHOD_TEST_CAPABILITY: self.handle_test_capability,
			self.METHOD_GET_NAME: self.handle_get_name,
			self.METHOD_GET_ACCOUNT_DATA: self.handle_get_account_data,
			self.METHOD_GET_PRIVATE_DATA: self.handle_get_private_data,
			self.METHOD_GET_PUBLIC_DATA: self.handle_get_public_data,
			self.METHOD_GET_MULTIPLE_PUBLIC_DATA: self.handle_get_multiple_public_data,
			self.METHOD_UPDATE_ACCOUNT_NAME: self.handle_update_account_name,
			self.METHOD_UPDATE_ACCOUNT_EMAIL: self.handle_update_account_email,
			self.METHOD_UPDATE_CUSTOM_DATA: self.handle_update_custom_data,
			self.METHOD_FIND_BY_NAME_REGEX: self.handle_find_by_name_regex,
			self.METHOD_UPDATE_ACCOUNT_EXPIRY_DATE: self.handle_update_account_expiry_date,
			self.METHOD_UPDATE_ACCOUNT_EFFECTIVE_DATE: self.handle_update_account_effective_date,
			self.METHOD_UPDATE_STATUS: self.handle_update_status,
			self.METHOD_GET_STATUS: self.handle_get_status,
			self.METHOD_GET_LAST_CONNECTION_STATS: self.handle_get_last_connection_stats,
			self.METHOD_RESET_PASSWORD: self.handle_reset_password,
			self.METHOD_CREATE_ACCOUNT_WITH_CUSTOM_DATA: self.handle_create_account_with_custom_data,
			self.METHOD_RETRIEVE_ACCOUNT: self.handle_retrieve_account,
			self.METHOD_UPDATE_ACCOUNT: self.handle_update_account,
			self.METHOD_CHANGE_PASSWORD_BY_GUEST: self.handle_change_password_by_guest,
			self.METHOD_FIND_BY_NAME_LIKE: self.handle_find_by_name_like,
			self.METHOD_CUSTOM_CREATE_ACCOUNT: self.handle_custom_create_account,
			self.METHOD_NINTENDO_CREATE_ACCOUNT: self.handle_nintendo_create_account,
			self.METHOD_LOOKUP_OR_CREATE_ACCOUNT: self.handle_lookup_or_create_account,
			self.METHOD_DISCONNECT_PRINCIPAL: self.handle_disconnect_principal,
			self.METHOD_DISCONNECT_ALL_PRINCIPALS: self.handle_disconnect_all_principals,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on AccountServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_create_account(self, client, input, output):
		logger.info("AccountServer.create_account()")
		#--- request ---
		name = input.string()
		key = input.string()
		groups = input.u32()
		email = input.string()
		response = await self.create_account(client, name, key, groups, email)
		
		#--- response ---
		if not isinstance(response, common.Result):
			raise RuntimeError("Expected common.Result, got %s" %response.__class__.__name__)
		output.result(response)
	
	async def handle_delete_account(self, client, input, output):
		logger.info("AccountServer.delete_account()")
		#--- request ---
		pid = input.pid()
		await self.delete_account(client, pid)
	
	async def handle_disable_account(self, client, input, output):
		logger.info("AccountServer.disable_account()")
		#--- request ---
		pid = input.pid()
		until = input.datetime()
		message = input.string()
		response = await self.disable_account(client, pid, until, message)
		
		#--- response ---
		if not isinstance(response, common.Result):
			raise RuntimeError("Expected common.Result, got %s" %response.__class__.__name__)
		output.result(response)
	
	async def handle_change_password(self, client, input, output):
		logger.info("AccountServer.change_password()")
		#--- request ---
		new_key = input.string()
		response = await self.change_password(client, new_key)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_test_capability(self, client, input, output):
		logger.info("AccountServer.test_capability()")
		#--- request ---
		capability = input.u32()
		response = await self.test_capability(client, capability)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_get_name(self, client, input, output):
		logger.info("AccountServer.get_name()")
		#--- request ---
		pid = input.pid()
		response = await self.get_name(client, pid)
		
		#--- response ---
		if not isinstance(response, str):
			raise RuntimeError("Expected str, got %s" %response.__class__.__name__)
		output.string(response)
	
	async def handle_get_account_data(self, client, input, output):
		logger.info("AccountServer.get_account_data()")
		#--- request ---
		response = await self.get_account_data(client)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.add(response.data)
	
	async def handle_get_private_data(self, client, input, output):
		logger.info("AccountServer.get_private_data()")
		#--- request ---
		response = await self.get_private_data(client)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.bool(response.result)
		output.anydata(response.data)
	
	async def handle_get_public_data(self, client, input, output):
		logger.info("AccountServer.get_public_data()")
		#--- request ---
		pid = input.pid()
		response = await self.get_public_data(client, pid)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.bool(response.result)
		output.anydata(response.data)
	
	async def handle_get_multiple_public_data(self, client, input, output):
		logger.info("AccountServer.get_multiple_public_data()")
		#--- request ---
		pids = input.list(input.pid)
		response = await self.get_multiple_public_data(client, pids)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.bool(response.result)
		output.list(response.data, output.anydata)
	
	async def handle_update_account_name(self, client, input, output):
		logger.info("AccountServer.update_account_name()")
		#--- request ---
		name = input.string()
		response = await self.update_account_name(client, name)
		
		#--- response ---
		if not isinstance(response, common.Result):
			raise RuntimeError("Expected common.Result, got %s" %response.__class__.__name__)
		output.result(response)
	
	async def handle_update_account_email(self, client, input, output):
		logger.info("AccountServer.update_account_email()")
		#--- request ---
		email = input.string()
		response = await self.update_account_email(client, email)
		
		#--- response ---
		if not isinstance(response, common.Result):
			raise RuntimeError("Expected common.Result, got %s" %response.__class__.__name__)
		output.result(response)
	
	async def handle_update_custom_data(self, client, input, output):
		logger.info("AccountServer.update_custom_data()")
		#--- request ---
		public_data = input.anydata()
		private_data = input.anydata()
		response = await self.update_custom_data(client, public_data, private_data)
		
		#--- response ---
		if not isinstance(response, common.Result):
			raise RuntimeError("Expected common.Result, got %s" %response.__class__.__name__)
		output.result(response)
	
	async def handle_find_by_name_regex(self, client, input, output):
		logger.info("AccountServer.find_by_name_regex()")
		#--- request ---
		groups = input.u32()
		regex = input.string()
		range = input.extract(common.ResultRange)
		response = await self.find_by_name_regex(client, groups, regex, range)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_update_account_expiry_date(self, client, input, output):
		logger.info("AccountServer.update_account_expiry_date()")
		#--- request ---
		pid = input.pid()
		expiry = input.datetime()
		message = input.string()
		await self.update_account_expiry_date(client, pid, expiry, message)
	
	async def handle_update_account_effective_date(self, client, input, output):
		logger.info("AccountServer.update_account_effective_date()")
		#--- request ---
		pid = input.pid()
		effective_from = input.datetime()
		message = input.string()
		await self.update_account_effective_date(client, pid, effective_from, message)
	
	async def handle_update_status(self, client, input, output):
		logger.info("AccountServer.update_status()")
		#--- request ---
		status = input.string()
		await self.update_status(client, status)
	
	async def handle_get_status(self, client, input, output):
		logger.info("AccountServer.get_status()")
		#--- request ---
		pid = input.pid()
		response = await self.get_status(client, pid)
		
		#--- response ---
		if not isinstance(response, str):
			raise RuntimeError("Expected str, got %s" %response.__class__.__name__)
		output.string(response)
	
	async def handle_get_last_connection_stats(self, client, input, output):
		logger.info("AccountServer.get_last_connection_stats()")
		#--- request ---
		pid = input.pid()
		response = await self.get_last_connection_stats(client, pid)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['last_session_login', 'last_session_logout', 'current_session_login']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.datetime(response.last_session_login)
		output.datetime(response.last_session_logout)
		output.datetime(response.current_session_login)
	
	async def handle_reset_password(self, client, input, output):
		logger.info("AccountServer.reset_password()")
		#--- request ---
		response = await self.reset_password(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_create_account_with_custom_data(self, client, input, output):
		logger.info("AccountServer.create_account_with_custom_data()")
		#--- request ---
		name = input.string()
		key = input.string()
		groups = input.u32()
		email = input.string()
		public_data = input.anydata()
		private_data = input.anydata()
		await self.create_account_with_custom_data(client, name, key, groups, email, public_data, private_data)
	
	async def handle_retrieve_account(self, client, input, output):
		logger.info("AccountServer.retrieve_account()")
		#--- request ---
		response = await self.retrieve_account(client)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['account_data', 'public_data', 'private_data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.add(response.account_data)
		output.anydata(response.public_data)
		output.anydata(response.private_data)
	
	async def handle_update_account(self, client, input, output):
		logger.info("AccountServer.update_account()")
		#--- request ---
		key = input.string()
		email = input.string()
		public_data = input.anydata()
		private_data = input.anydata()
		await self.update_account(client, key, email, public_data, private_data)
	
	async def handle_change_password_by_guest(self, client, input, output):
		logger.info("AccountServer.change_password_by_guest()")
		#--- request ---
		name = input.string()
		email = input.string()
		key = input.string()
		await self.change_password_by_guest(client, name, email, key)
	
	async def handle_find_by_name_like(self, client, input, output):
		logger.info("AccountServer.find_by_name_like()")
		#--- request ---
		groups = input.u32()
		like = input.string()
		range = input.extract(common.ResultRange)
		response = await self.find_by_name_like(client, groups, like, range)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_custom_create_account(self, client, input, output):
		logger.info("AccountServer.custom_create_account()")
		#--- request ---
		name = input.string()
		key = input.string()
		groups = input.u32()
		email = input.string()
		auth_data = input.anydata()
		response = await self.custom_create_account(client, name, key, groups, email, auth_data)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.pid(response)
	
	async def handle_nintendo_create_account(self, client, input, output):
		logger.info("AccountServer.nintendo_create_account()")
		#--- request ---
		name = input.string()
		key = input.string()
		groups = input.u32()
		email = input.string()
		auth_data = input.anydata()
		response = await self.nintendo_create_account(client, name, key, groups, email, auth_data)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['pid', 'pid_hmac']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.pid(response.pid)
		output.string(response.pid_hmac)
	
	async def handle_lookup_or_create_account(self, client, input, output):
		logger.info("AccountServer.lookup_or_create_account()")
		#--- request ---
		name = input.string()
		key = input.string()
		groups = input.u32()
		email = input.string()
		auth_data = input.anydata()
		response = await self.lookup_or_create_account(client, name, key, groups, email, auth_data)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.pid(response)
	
	async def handle_disconnect_principal(self, client, input, output):
		logger.info("AccountServer.disconnect_principal()")
		#--- request ---
		pid = input.pid()
		response = await self.disconnect_principal(client, pid)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_disconnect_all_principals(self, client, input, output):
		logger.info("AccountServer.disconnect_all_principals()")
		#--- request ---
		response = await self.disconnect_all_principals(client)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def create_account(self, *args):
		logger.warning("AccountServer.create_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_account(self, *args):
		logger.warning("AccountServer.delete_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def disable_account(self, *args):
		logger.warning("AccountServer.disable_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def change_password(self, *args):
		logger.warning("AccountServer.change_password not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def test_capability(self, *args):
		logger.warning("AccountServer.test_capability not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_name(self, *args):
		logger.warning("AccountServer.get_name not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_account_data(self, *args):
		logger.warning("AccountServer.get_account_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_private_data(self, *args):
		logger.warning("AccountServer.get_private_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_public_data(self, *args):
		logger.warning("AccountServer.get_public_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_multiple_public_data(self, *args):
		logger.warning("AccountServer.get_multiple_public_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_account_name(self, *args):
		logger.warning("AccountServer.update_account_name not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_account_email(self, *args):
		logger.warning("AccountServer.update_account_email not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_custom_data(self, *args):
		logger.warning("AccountServer.update_custom_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def find_by_name_regex(self, *args):
		logger.warning("AccountServer.find_by_name_regex not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_account_expiry_date(self, *args):
		logger.warning("AccountServer.update_account_expiry_date not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_account_effective_date(self, *args):
		logger.warning("AccountServer.update_account_effective_date not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_status(self, *args):
		logger.warning("AccountServer.update_status not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_status(self, *args):
		logger.warning("AccountServer.get_status not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_last_connection_stats(self, *args):
		logger.warning("AccountServer.get_last_connection_stats not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def reset_password(self, *args):
		logger.warning("AccountServer.reset_password not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def create_account_with_custom_data(self, *args):
		logger.warning("AccountServer.create_account_with_custom_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def retrieve_account(self, *args):
		logger.warning("AccountServer.retrieve_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_account(self, *args):
		logger.warning("AccountServer.update_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def change_password_by_guest(self, *args):
		logger.warning("AccountServer.change_password_by_guest not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def find_by_name_like(self, *args):
		logger.warning("AccountServer.find_by_name_like not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def custom_create_account(self, *args):
		logger.warning("AccountServer.custom_create_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def nintendo_create_account(self, *args):
		logger.warning("AccountServer.nintendo_create_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def lookup_or_create_account(self, *args):
		logger.warning("AccountServer.lookup_or_create_account not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def disconnect_principal(self, *args):
		logger.warning("AccountServer.disconnect_principal not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def disconnect_all_principals(self, *args):
		logger.warning("AccountServer.disconnect_all_principals not implemented")
		raise common.RMCError("Core::NotImplemented")





# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class NotificationEvent(common.Structure):
	def __init__(self):
		super().__init__()
		self.pid = None
		self.type = None
		self.param1 = 0
		self.param2 = 0
		self.text = ""
		self.param3 = 0
		self.map = {}
	
	def max_version(self, settings):
		version = 0
		if settings["nex.version"] >= 40000:
			version = 1
		return version
	
	def check_required(self, settings, version):
		for field in ['pid', 'type']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
		if settings["nex.version"] >= 30500:
			pass
		if settings["nex.version"] >= 40000:
			if version >= 1:
				pass
	
	def load(self, stream, version):
		self.pid = stream.pid()
		self.type = stream.u32()
		self.param1 = stream.pid()
		self.param2 = stream.pid()
		self.text = stream.string()
		if stream.settings["nex.version"] >= 30500:
			self.param3 = stream.pid()
		if stream.settings["nex.version"] >= 40000:
			if version >= 1:
				self.map = stream.map(stream.string, stream.variant)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.pid(self.pid)
		stream.u32(self.type)
		stream.pid(self.param1)
		stream.pid(self.param2)
		stream.string(self.text)
		if stream.settings["nex.version"] >= 30500:
			stream.pid(self.param3)
		if stream.settings["nex.version"] >= 40000:
			if version >= 1:
				stream.map(self.map, stream.string, stream.variant)


class NotificationProtocol:
	NORESPONSE = True
	
	METHOD_PROCESS_NOTIFICATION_EVENT = 1
	
	PROTOCOL_ID = 0xE


class NotificationClient(NotificationProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def process_notification_event(self, event):
		logger.info("NotificationClient.process_notification_event()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(event)
		await self.client.request(self.PROTOCOL_ID, self.METHOD_PROCESS_NOTIFICATION_EVENT, stream.get(), True)


class NotificationServer(NotificationProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_PROCESS_NOTIFICATION_EVENT: self.handle_process_notification_event,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on NotificationServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_process_notification_event(self, client, input, output):
		logger.info("NotificationServer.process_notification_event()")
		#--- request ---
		event = input.extract(NotificationEvent)
		await self.process_notification_event(client, event)
	
	async def process_notification_event(self, *args):
		logger.warning("NotificationServer.process_notification_event not implemented")
		raise common.RMCError("Core::NotImplemented")





from nintendo.nex import prudp, common, streams
from anynet import util
import contextlib
import struct
import anyio

import logging
logger = logging.getLogger(__name__)


class RMCResponse:
	pass


class RMCMessage:
	REQUEST = 0
	RESPONSE = 1
	
	def __init__(self, settings):
		self.settings = settings
		
		self.mode = RMCMessage.REQUEST
		self.protocol = None
		self.method = None
		self.call_id = 0
		self.error = -1
		self.body = b""
		
	@staticmethod
	def prepare(settings, mode, protocol, method, call_id, body):
		inst = RMCMessage(settings)
		inst.mode = mode
		inst.protocol = protocol
		inst.method = method
		inst.call_id = call_id
		inst.body = body
		return inst
		
	@staticmethod
	def request(settings, protocol, method, call_id, body):
		return RMCMessage.prepare(
			settings, RMCMessage.REQUEST, protocol, method, call_id, body
		)
		
	@staticmethod
	def response(settings, protocol, method, call_id, body):
		return RMCMessage.prepare(
			settings, RMCMessage.RESPONSE, protocol, method, call_id, body
		)
		
	@staticmethod
	def error(settings, protocol, method, call_id, error):
		inst = RMCMessage(settings)
		inst.mode = RMCMessage.RESPONSE
		inst.protocol = protocol
		inst.method = method
		inst.call_id = call_id
		inst.error = error
		return inst
	
	@staticmethod
	def parse(settings, data):
		inst = RMCMessage(settings)
		inst.decode(data)
		return inst
		
	def encode(self):
		stream = streams.StreamOut(self.settings)
		
		flag = 0x80 if self.mode == self.REQUEST else 0
		if self.protocol < 0x80:
			stream.u8(self.protocol | flag)
		else:
			stream.u8(0x7F | flag)
			stream.u16(self.protocol)
		
		if self.mode == self.REQUEST:
			stream.u32(self.call_id)
			stream.u32(self.method)
			stream.write(self.body)
		else:
			if self.error != -1 and self.error & 0x80000000:
				stream.bool(False)
				stream.u32(self.error)
				stream.u32(self.call_id)
			else:
				stream.bool(True)
				stream.u32(self.call_id)
				stream.u32(self.method | 0x8000)
				stream.write(self.body)
		return struct.pack("I", stream.size()) + stream.get()
	
	def decode(self, data):
		stream = streams.StreamIn(data, self.settings)
		
		length = stream.u32()
		if length != stream.size() - 4:
			raise ValueError("RMC message has unexpected size")
		
		protocol = stream.u8()
		self.protocol = protocol & ~0x80
		if self.protocol == 0x7F:
			self.protocol = stream.u16()
		
		if protocol & 0x80:
			self.mode = self.REQUEST
			self.call_id = stream.u32()
			self.method = stream.u32()
			self.body = stream.readall()
		else:
			self.mode = self.RESPONSE
			if stream.bool():
				self.call_id = stream.u32()
				self.method = stream.u32() & ~0x8000
				self.body = stream.readall()
			else:
				self.error = stream.u32()
				self.call_id = stream.u32()
				if not stream.eof():
					raise ValueError("RMC error message is bigger than expected")


class RMCClient:
	def __init__(self, settings, client):
		self.settings = settings.copy()
		self.client = client
		self.call_id = 1
		
		if self.client.minor_version() >= 3:
			self.settings["nex.struct_header"] = True
		
		self.servers = {}
		self.requests = {}
		self.responses = {}
		
		self.closed = False

	async def __aenter__(self): return self
	async def __aexit__(self, typ, val, tb):
		await self.cleanup()
	
	def register_server(self, server):
		if server.PROTOCOL_ID in self.servers:
			raise ValueError("Server with protocol id %i already exists" %server.PROTOCOL_ID)
		self.servers[server.PROTOCOL_ID] = server
	
	async def cleanup(self):
		if not self.closed:
			self.closed = True
			for event in self.requests.values():
				event.set()
			
			for server in self.servers.values():
				await server.logout(self)
	
	async def close(self):
		if not self.closed:
			await self.cleanup()
			await self.client.close()
	
	async def disconnect(self):
		if not self.closed:
			await self.cleanup()
			await self.client.disconnect()
	
	async def start(self, servers):
		for server in servers:
			self.register_server(server)
		
		while not self.closed:
			try:
				data = await self.client.recv()
			except anyio.EndOfStream:
				logger.info("Connection was closed")
				await self.cleanup()
				return
			
			message = RMCMessage.parse(self.settings, data)
			if message.mode == RMCMessage.REQUEST:
				logger.debug(
					"Received RMC request: protocol=%i method=%i call=%i",
					message.protocol, message.method, message.call_id
				)
				await self.handle_request(message)
			else:
				logger.debug(
					"Received RMC response: protocol=%i method=%s call=%i",
					message.protocol, message.method, message.call_id
				)
				if message.call_id in self.requests:
					self.responses[message.call_id] = message
					event = self.requests.pop(message.call_id)
					event.set()
				else:
					logger.warning("RMC response has invalid call id")
	
	async def handle_request(self, request):
		input = streams.StreamIn(request.body, self.settings)
		output = streams.StreamOut(self.settings)
		
		result = common.Result()
		if request.protocol in self.servers:
			try:
				await self.servers[request.protocol].handle(self, request.method, input, output)
			except common.RMCError as e:
				logger.warning("RMC failed: %s" %e)
				result = e.result()
			except Exception as e:
				logger.exception("Exception occurred while handling a method call")
				
				if isinstance(e, TypeError): result = common.Result.error("PythonCore::TypeError")
				elif isinstance(e, IndexError): result = common.Result.error("PythonCore::IndexError")
				elif isinstance(e, MemoryError): result = common.Result.error("PythonCore::MemoryError")
				elif isinstance(e, KeyError): result = common.Result.error("PythonCore::KeyError")
				else: result = common.Result.error("PythonCore::Exception")
			except anyio.ExceptionGroup as e:
				logger.exception("Multiple exceptions occurred while handling a method call")
				
				filtered = []
				for exc in e.exceptions:
					if not isinstance(exc, Exception):
						raise
				
				result = common.Result.error("PythonCore::Exception")
			
			if getattr(self.servers[request.protocol], "NORESPONSE", False):
				return
		else:
			logger.warning("Received RMC request with unimplemented protocol id: %i", request.protocol)
			result = common.Result.error("Core::NotImplemented")
		
		if result.is_success():
			response = RMCMessage.response(
				self.settings, request.protocol, request.method,
				request.call_id, output.get()
			)
		else:
			response = RMCMessage.error(
				self.settings, request.protocol, request.method,
				request.call_id, result.code()
			)
		await self.client.send(response.encode())
	
	async def request(self, protocol, method, body, noresponse=False):
		if self.closed:
			raise RuntimeError("RMC connection is closed")
		
		call_id = self.call_id
		self.call_id = (self.call_id + 1) & 0xFFFFFFFF
		
		if not noresponse:
			event = anyio.Event()
			self.requests[call_id] = event
		
		message = RMCMessage.request(self.settings, protocol, method, call_id, body)
		await self.client.send(message.encode())
		
		if not noresponse:
			await event.wait()
			
			if self.closed:
				raise RuntimeError("RMC connection is closed")
			
			message = self.responses.pop(call_id)
			if message.error != -1:
				raise common.RMCError(message.error)
			return message.body
	
	def pid(self): return self.client.pid()
	
	def local_address(self):
		return self.client.local_address()
	def remote_address(self):
		return self.client.remote_address()
	
	def local_sid(self):
		return self.client.local_sid()
	def remote_sid(self):
		return self.client.remote_sid()


@contextlib.asynccontextmanager
async def connect(settings, host, port, vport=1, context=None, credentials=None, servers=[]):
	logger.debug("Connecting RMC client to %s:%i:%i", host, port, vport)
	async with prudp.connect(settings, host, port, vport, 10, context, credentials) as client:
		client = RMCClient(settings, client)
		async with client:
			async with util.create_task_group() as group:
				group.start_soon(client.start, servers)
				yield client
	logger.debug("RMC client is closed")

@contextlib.asynccontextmanager
async def serve(settings, servers, host="", port=0, vport=1, context=None, key=None):
	async def handle(client):
		host, port = client.remote_address()
		logger.debug("New RMC connection: %s:%i", host, port)
		
		client = RMCClient(settings, client)
		async with client:
			await client.start(servers)
	
	logger.info("Starting RMC server at %s:%i:%i", host, port, vport)
	async with prudp.serve(handle, settings, host, port, vport, 10, context, key):
		yield
	logger.info("RMC server is closed")

@contextlib.asynccontextmanager
async def serve_on_transport(settings, servers, transport, port, key=None):
	async def handle(client):
		host, port = client.remote_address()
		logger.debug("New RMC connection: %s:%i", host, port)
		
		client = RMCClient(settings, client)
		async with client:
			await client.start(servers)
	
	logger.info("Starting RMC server at PRUDP port %i", port)
	async with transport.serve(handle, port, 10, key):
		yield
	logger.info("RMC server is closed")

serve_prudp = serve_on_transport




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class SubscriberProtocol:
	METHOD_HELLO = 1
	METHOD_POST_CONTENT = 2
	METHOD_GET_CONTENT = 3
	METHOD_FOLLOW = 4
	METHOD_UNFOLLOW_ALL_AND_FOLLOW = 5
	METHOD_UNFOLLOW = 6
	METHOD_GET_FOLLOWING = 7
	METHOD_GET_FOLLOWER = 8
	METHOD_GET_NUM_FOLLOWERS = 9
	METHOD_GET_TIMELINE = 10
	METHOD_DELETE_CONTENT = 11
	METHOD_GET_CONTENT_MULTI = 12
	METHOD_UPDATE_USER_STATUS = 13
	METHOD_GET_FRIEND_USER_STATUSES = 14
	METHOD_GET_USER_STATUSES = 15
	
	PROTOCOL_ID = 0x79


class SubscriberClient(SubscriberProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	


class SubscriberServer(SubscriberProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_HELLO: self.handle_hello,
			self.METHOD_POST_CONTENT: self.handle_post_content,
			self.METHOD_GET_CONTENT: self.handle_get_content,
			self.METHOD_FOLLOW: self.handle_follow,
			self.METHOD_UNFOLLOW_ALL_AND_FOLLOW: self.handle_unfollow_all_and_follow,
			self.METHOD_UNFOLLOW: self.handle_unfollow,
			self.METHOD_GET_FOLLOWING: self.handle_get_following,
			self.METHOD_GET_FOLLOWER: self.handle_get_follower,
			self.METHOD_GET_NUM_FOLLOWERS: self.handle_get_num_followers,
			self.METHOD_GET_TIMELINE: self.handle_get_timeline,
			self.METHOD_DELETE_CONTENT: self.handle_delete_content,
			self.METHOD_GET_CONTENT_MULTI: self.handle_get_content_multi,
			self.METHOD_UPDATE_USER_STATUS: self.handle_update_user_status,
			self.METHOD_GET_FRIEND_USER_STATUSES: self.handle_get_friend_user_statuses,
			self.METHOD_GET_USER_STATUSES: self.handle_get_user_statuses,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on SubscriberServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_hello(self, client, input, output):
		logger.warning("SubscriberServer.hello is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_post_content(self, client, input, output):
		logger.warning("SubscriberServer.post_content is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_content(self, client, input, output):
		logger.warning("SubscriberServer.get_content is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_follow(self, client, input, output):
		logger.warning("SubscriberServer.follow is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_unfollow_all_and_follow(self, client, input, output):
		logger.warning("SubscriberServer.unfollow_all_and_follow is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_unfollow(self, client, input, output):
		logger.warning("SubscriberServer.unfollow is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_following(self, client, input, output):
		logger.warning("SubscriberServer.get_following is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_follower(self, client, input, output):
		logger.warning("SubscriberServer.get_follower is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_num_followers(self, client, input, output):
		logger.warning("SubscriberServer.get_num_followers is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_timeline(self, client, input, output):
		logger.warning("SubscriberServer.get_timeline is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_delete_content(self, client, input, output):
		logger.warning("SubscriberServer.delete_content is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_content_multi(self, client, input, output):
		logger.warning("SubscriberServer.get_content_multi is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_update_user_status(self, client, input, output):
		logger.warning("SubscriberServer.update_user_status is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_friend_user_statuses(self, client, input, output):
		logger.warning("SubscriberServer.get_friend_user_statuses is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_user_statuses(self, client, input, output):
		logger.warning("SubscriberServer.get_user_statuses is not supported")
		raise common.RMCError("Core::NotImplemented")





from nintendo.nex import rmc, authentication, kerberos, common
from anynet import tls
import contextlib

import logging
logger = logging.getLogger(__name__)


class LoginResult:
	def __init__(self, pid, ticket, source_key, secure_station):
		self.pid = pid
		self.ticket = ticket
		self.source_key = source_key
		self.secure_station = secure_station


class BackEndClient:
	def __init__(self, settings, client, host, port):
		self.settings = settings
		self.auth_client = client
		self.auth_host = host
		self.auth_port = port
		
		if self.settings["nex.version"] < 40000:
			self.auth_proto = authentication.AuthenticationClient(client)
		else:
			self.auth_proto = authentication.AuthenticationClientNX(client)
		
		if self.settings["kerberos.key_derivation"] == 0:
			self.key_derivation = kerberos.KeyDerivationOld(65000, 1024)
		else:
			self.key_derivation = kerberos.KeyDerivationNew(1, 1)
	
	@contextlib.asynccontextmanager
	async def login(self, username, password=None, auth_info=None, servers=[]):
		if self.settings["nex.version"] < 40000:
			result = await self.login_old(username, auth_info)
		elif self.settings["nex.version"] < 40400:
			result = await self.login_switch(username, auth_info)
		else:
			result = await self.login_with_param(username, auth_info)
		
		secure_station = result.secure_station
		
		kerberos_key = result.source_key
		if not kerberos_key:
			if password is None:
				raise ValueError("A password is required for this account")
			
			# Derive kerberos key from password
			kerberos_key = self.key_derivation.derive_key(
				password.encode(), result.pid
			)
		
		# Decrypt ticket from login response
		ticket = kerberos.ClientTicket.decrypt(result.ticket, kerberos_key, self.settings)
		if ticket.target != secure_station["PID"]:
			# Request ticket for secure server
			response = await self.auth_proto.request_ticket(
				result.pid, secure_station["PID"]
			)
			
			# Check for errors and decrypt ticket
			response.result.raise_if_error()
			ticket = kerberos.ClientTicket.decrypt(response.ticket, kerberos_key, self.settings)
		
		creds = kerberos.Credentials(ticket, result.pid, secure_station["CID"])
		
		# The secure server may reside at the same
		# address as the authentication server
		host = secure_station["address"]
		port = secure_station["port"]
		if host == "0.0.0.1":
			host, port = self.auth_host, self.auth_port

		# Connect to secure server
		stream_id = secure_station["sid"]
		
		context = tls.TLSContext()
		async with rmc.connect(self.settings, host, port, stream_id, context, creds, servers) as client:
			yield client
	
	async def login_old(self, username, auth_info):
		if auth_info:
			response = await self.auth_proto.login_ex(username, auth_info)
		else:
			response = await self.auth_proto.login(username)
		response.result.raise_if_error()
		return LoginResult(
			response.pid, response.ticket, None,
			response.connection_data.main_station
		)
	
	async def login_switch(self, username, auth_info):
		if auth_info:
			response = await self.auth_proto.validate_and_request_ticket_with_custom_data(
				username, auth_info
			)
			response.result.raise_if_error()
			return LoginResult(
				response.pid, response.ticket, bytes.fromhex(response.source_key),
				response.connection_data.main_station
			)
		else:
			response = await self.auth_proto.validate_and_request_ticket(username)
			response.result.raise_if_error()
			return LoginResult(
				response.pid, response.ticket, None,
				response.connection_data.main_station
			)
		
	async def login_with_param(self, username, auth_info):
		param = authentication.ValidateAndRequestTicketParam()
		param.username = username
		if auth_info:
			param.data = auth_info
		else:
			param.data = common.NullData()
		param.nex_version = self.settings["nex.version"]
		param.client_version = self.settings["nex.client_version"]
		
		response = await self.auth_proto.validate_and_request_ticket_with_param(param)
		return LoginResult(
			response.pid, response.ticket,
			bytes.fromhex(response.source_key),
			response.server_url
		)
	
	def login_guest(self):
		return self.login("guest", "MMQea3n!fsik")


@contextlib.asynccontextmanager
async def connect(settings, host, port):
	context = tls.TLSContext()
	async with rmc.connect(settings, host, port, context=context) as client:
		yield BackEndClient(settings, client, host, port)




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class UniqueIdInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.unique_id = 0
		self.password = 0
	
	def check_required(self, settings, version):
		pass
	
	def load(self, stream, version):
		self.unique_id = stream.u64()
		self.password = stream.u64()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u64(self.unique_id)
		stream.u64(self.password)


class UtilityProtocol:
	METHOD_ACQUIRE_NEX_UNIQUE_ID = 1
	METHOD_ACQUIRE_NEX_UNIQUE_ID_WITH_PASSWORD = 2
	METHOD_ASSOCIATE_NEX_UNIQUE_ID_WITH_MY_PRINCIPAL_ID = 3
	METHOD_ASSOCIATE_NEX_UNIQUE_IDS_WITH_MY_PRINCIPAL_ID = 4
	METHOD_GET_ASSOCIATED_NEX_UNIQUE_ID_WITH_MY_PRINCIPAL_ID = 5
	METHOD_GET_ASSOCIATED_NEX_UNIQUE_IDS_WITH_MY_PRINCIPAL_ID = 6
	METHOD_GET_INTEGER_SETTINGS = 7
	METHOD_GET_STRING_SETTINGS = 8
	
	PROTOCOL_ID = 0x6E


class UtilityClient(UtilityProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def acquire_nex_unique_id(self):
		logger.info("UtilityClient.acquire_nex_unique_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_ACQUIRE_NEX_UNIQUE_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		unique_id = stream.u64()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.acquire_nex_unique_id -> done")
		return unique_id
	
	async def acquire_nex_unique_id_with_password(self):
		logger.info("UtilityClient.acquire_nex_unique_id_with_password()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_ACQUIRE_NEX_UNIQUE_ID_WITH_PASSWORD, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.extract(UniqueIdInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.acquire_nex_unique_id_with_password -> done")
		return info
	
	async def associate_nex_unique_id_with_my_principal_id(self, info):
		logger.info("UtilityClient.associate_nex_unique_id_with_my_principal_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(info)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_ASSOCIATE_NEX_UNIQUE_ID_WITH_MY_PRINCIPAL_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.associate_nex_unique_id_with_my_principal_id -> done")
	
	async def associate_nex_unique_ids_with_my_principal_id(self, infos):
		logger.info("UtilityClient.associate_nex_unique_ids_with_my_principal_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(infos, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_ASSOCIATE_NEX_UNIQUE_IDS_WITH_MY_PRINCIPAL_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.associate_nex_unique_ids_with_my_principal_id -> done")
	
	async def get_associated_nex_unique_id_with_my_principal_id(self):
		logger.info("UtilityClient.get_associated_nex_unique_id_with_my_principal_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_ASSOCIATED_NEX_UNIQUE_ID_WITH_MY_PRINCIPAL_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.extract(UniqueIdInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.get_associated_nex_unique_id_with_my_principal_id -> done")
		return info
	
	async def get_associated_nex_unique_ids_with_my_principal_id(self):
		logger.info("UtilityClient.get_associated_nex_unique_ids_with_my_principal_id()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_ASSOCIATED_NEX_UNIQUE_IDS_WITH_MY_PRINCIPAL_ID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		infos = stream.list(UniqueIdInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.get_associated_nex_unique_ids_with_my_principal_id -> done")
		return infos
	
	async def get_integer_settings(self, index):
		logger.info("UtilityClient.get_integer_settings()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(index)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_INTEGER_SETTINGS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		settings = stream.map(stream.u16, stream.s32)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.get_integer_settings -> done")
		return settings
	
	async def get_string_settings(self, index):
		logger.info("UtilityClient.get_string_settings()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(index)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_STRING_SETTINGS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		settings = stream.map(stream.u16, stream.string)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("UtilityClient.get_string_settings -> done")
		return settings


class UtilityServer(UtilityProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_ACQUIRE_NEX_UNIQUE_ID: self.handle_acquire_nex_unique_id,
			self.METHOD_ACQUIRE_NEX_UNIQUE_ID_WITH_PASSWORD: self.handle_acquire_nex_unique_id_with_password,
			self.METHOD_ASSOCIATE_NEX_UNIQUE_ID_WITH_MY_PRINCIPAL_ID: self.handle_associate_nex_unique_id_with_my_principal_id,
			self.METHOD_ASSOCIATE_NEX_UNIQUE_IDS_WITH_MY_PRINCIPAL_ID: self.handle_associate_nex_unique_ids_with_my_principal_id,
			self.METHOD_GET_ASSOCIATED_NEX_UNIQUE_ID_WITH_MY_PRINCIPAL_ID: self.handle_get_associated_nex_unique_id_with_my_principal_id,
			self.METHOD_GET_ASSOCIATED_NEX_UNIQUE_IDS_WITH_MY_PRINCIPAL_ID: self.handle_get_associated_nex_unique_ids_with_my_principal_id,
			self.METHOD_GET_INTEGER_SETTINGS: self.handle_get_integer_settings,
			self.METHOD_GET_STRING_SETTINGS: self.handle_get_string_settings,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on UtilityServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_acquire_nex_unique_id(self, client, input, output):
		logger.info("UtilityServer.acquire_nex_unique_id()")
		#--- request ---
		response = await self.acquire_nex_unique_id(client)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.u64(response)
	
	async def handle_acquire_nex_unique_id_with_password(self, client, input, output):
		logger.info("UtilityServer.acquire_nex_unique_id_with_password()")
		#--- request ---
		response = await self.acquire_nex_unique_id_with_password(client)
		
		#--- response ---
		if not isinstance(response, UniqueIdInfo):
			raise RuntimeError("Expected UniqueIdInfo, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_associate_nex_unique_id_with_my_principal_id(self, client, input, output):
		logger.info("UtilityServer.associate_nex_unique_id_with_my_principal_id()")
		#--- request ---
		info = input.extract(UniqueIdInfo)
		await self.associate_nex_unique_id_with_my_principal_id(client, info)
	
	async def handle_associate_nex_unique_ids_with_my_principal_id(self, client, input, output):
		logger.info("UtilityServer.associate_nex_unique_ids_with_my_principal_id()")
		#--- request ---
		infos = input.list(UniqueIdInfo)
		await self.associate_nex_unique_ids_with_my_principal_id(client, infos)
	
	async def handle_get_associated_nex_unique_id_with_my_principal_id(self, client, input, output):
		logger.info("UtilityServer.get_associated_nex_unique_id_with_my_principal_id()")
		#--- request ---
		response = await self.get_associated_nex_unique_id_with_my_principal_id(client)
		
		#--- response ---
		if not isinstance(response, UniqueIdInfo):
			raise RuntimeError("Expected UniqueIdInfo, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_associated_nex_unique_ids_with_my_principal_id(self, client, input, output):
		logger.info("UtilityServer.get_associated_nex_unique_ids_with_my_principal_id()")
		#--- request ---
		response = await self.get_associated_nex_unique_ids_with_my_principal_id(client)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_integer_settings(self, client, input, output):
		logger.info("UtilityServer.get_integer_settings()")
		#--- request ---
		index = input.u32()
		response = await self.get_integer_settings(client, index)
		
		#--- response ---
		if not isinstance(response, dict):
			raise RuntimeError("Expected dict, got %s" %response.__class__.__name__)
		output.map(response, output.u16, output.s32)
	
	async def handle_get_string_settings(self, client, input, output):
		logger.info("UtilityServer.get_string_settings()")
		#--- request ---
		index = input.u32()
		response = await self.get_string_settings(client, index)
		
		#--- response ---
		if not isinstance(response, dict):
			raise RuntimeError("Expected dict, got %s" %response.__class__.__name__)
		output.map(response, output.u16, output.string)
	
	async def acquire_nex_unique_id(self, *args):
		logger.warning("UtilityServer.acquire_nex_unique_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def acquire_nex_unique_id_with_password(self, *args):
		logger.warning("UtilityServer.acquire_nex_unique_id_with_password not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def associate_nex_unique_id_with_my_principal_id(self, *args):
		logger.warning("UtilityServer.associate_nex_unique_id_with_my_principal_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def associate_nex_unique_ids_with_my_principal_id(self, *args):
		logger.warning("UtilityServer.associate_nex_unique_ids_with_my_principal_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_associated_nex_unique_id_with_my_principal_id(self, *args):
		logger.warning("UtilityServer.get_associated_nex_unique_id_with_my_principal_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_associated_nex_unique_ids_with_my_principal_id(self, *args):
		logger.warning("UtilityServer.get_associated_nex_unique_ids_with_my_principal_id not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_integer_settings(self, *args):
		logger.warning("UtilityServer.get_integer_settings not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_string_settings(self, *args):
		logger.warning("UtilityServer.get_string_settings not implemented")
		raise common.RMCError("Core::NotImplemented")





from anynet import tls, http
from nintendo import resources
from nintendo.switch import common

import logging
logger = logging.getLogger(__name__)


USER_AGENT = "NintendoSDK Firmware/%s (platform:NX; did:%016x; eid:lp1)"

LATEST_VERSION = 2120


class AtumnClient:
	def __init__(self, device_id):
		self.device_id = device_id
		
		self.request_callback = http.request
		
		ca = resources.certificate("Nintendo_Class_2_CA_G3.der")
		self.context = tls.TLSContext()
		self.context.set_authority(ca)
		
		self.host = "atumn.hac.lp1.d4c.nintendo.net"
		
		self.user_agent = USER_AGENT %(common.FIRMWARE_VERSIONS[LATEST_VERSION], self.device_id)
	
	def set_request_callback(self, callback): self.request_callback = callback
	
	def set_context(self, context): self.context = context
	def set_certificate(self, cert, key): self.context.set_certificate(cert, key)
	
	def set_host(self, host): self.host = host
	
	def set_system_version(self, version):
		if version not in common.FIRMWARE_VERSIONS:
			raise ValueError("Unknown system version: %i" %version)
		self.user_agent = USER_AGENT %(common.FIRMWARE_VERSIONS[version], self.device_id)
	
	async def request(self, req):
		req.headers["Host"] = self.host
		req.headers["Accept"] = "*/*"
		req.headers["User-Agent"] = self.user_agent
		
		response = await self.request_callback(self.host, req, self.context)
		response.raise_if_error()
		return response

	async def download_content_metadata(self, title_id, title_version, *, system_update=False):
		content_type = "s" if system_update else "a"

		req = http.HTTPRequest.head("/t/%c/%016x/%i" %(content_type, title_id, title_version))
		req.params = {
			"device_id": "%016x" %self.device_id
		}

		response = await self.request(req)
		content_id = response.headers["X-Nintendo-Content-ID"]

		req = http.HTTPRequest.get("/c/%s/%s" %(content_type, content_id))
		response = await self.request(req)
		return response.body
	
	async def download_content(self, content_id):
		req = http.HTTPRequest.get("/c/c/%s" %content_id)
		response = await self.request(req)
		return response.body




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class RankingOrderCalc:
	STANDARD = 0
	ORDINAL = 1


class RankingMode:
	GLOBAL = 0
	GLOBAL_AROUND_SELF = 1
	SELF = 4


class RankingStatFlags:
	RANKING_COUNT = 1
	TOTAL_SCORE = 2
	LOWEST_SCORE = 4
	HIGHEST_SCORE = 8
	AVERAGE_SCORE = 16
	ALL = 31


class RankingOrderParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.order_calc = 0
		self.group_index = 255
		self.group_num = 0
		self.time_scope = 2
		self.offset = 0
		self.count = 10
	
	def check_required(self, settings, version):
		pass
	
	def load(self, stream, version):
		self.order_calc = stream.u8()
		self.group_index = stream.u8()
		self.group_num = stream.u8()
		self.time_scope = stream.u8()
		self.offset = stream.u32()
		self.count = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u8(self.order_calc)
		stream.u8(self.group_index)
		stream.u8(self.group_num)
		stream.u8(self.time_scope)
		stream.u32(self.offset)
		stream.u8(self.count)


class RankingRankData(common.Structure):
	def __init__(self):
		super().__init__()
		self.pid = None
		self.unique_id = None
		self.rank = None
		self.category = None
		self.score = None
		self.groups = None
		self.param = None
		self.common_data = None
		self.update_time = None
	
	def check_required(self, settings, version):
		for field in ['pid', 'unique_id', 'rank', 'category', 'score', 'groups', 'param', 'common_data']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
		if settings["nex.version"] >= 40000:
			for field in ['update_time']:
				if getattr(self, field) is None:
					raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.pid = stream.pid()
		self.unique_id = stream.u64()
		self.rank = stream.u32()
		self.category = stream.u32()
		self.score = stream.u32()
		self.groups = stream.list(stream.u8)
		self.param = stream.u64()
		self.common_data = stream.buffer()
		if stream.settings["nex.version"] >= 40000:
			self.update_time = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.pid(self.pid)
		stream.u64(self.unique_id)
		stream.u32(self.rank)
		stream.u32(self.category)
		stream.u32(self.score)
		stream.list(self.groups, stream.u8)
		stream.u64(self.param)
		stream.buffer(self.common_data)
		if stream.settings["nex.version"] >= 40000:
			stream.datetime(self.update_time)


class RankingResult(common.Structure):
	def __init__(self):
		super().__init__()
		self.data = None
		self.total = None
		self.since_time = None
	
	def check_required(self, settings, version):
		for field in ['data', 'total', 'since_time']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.data = stream.list(RankingRankData)
		self.total = stream.u32()
		self.since_time = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.list(self.data, stream.add)
		stream.u32(self.total)
		stream.datetime(self.since_time)


class RankingCachedResult(RankingResult):
	def __init__(self):
		super().__init__()
		self.created_time = None
		self.expired_time = None
		self.max_length = None
	
	def check_required(self, settings, version):
		for field in ['created_time', 'expired_time', 'max_length']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.created_time = stream.datetime()
		self.expired_time = stream.datetime()
		self.max_length = stream.u8()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.datetime(self.created_time)
		stream.datetime(self.expired_time)
		stream.u8(self.max_length)
common.DataHolder.register(RankingCachedResult, "RankingCachedResult")


class RankingStats(common.Structure):
	def __init__(self):
		super().__init__()
		self.stats = None
	
	def check_required(self, settings, version):
		for field in ['stats']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.stats = stream.list(stream.double)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.list(self.stats, stream.double)


class RankingScoreData(common.Structure):
	def __init__(self):
		super().__init__()
		self.category = None
		self.score = None
		self.order = None
		self.update_mode = None
		self.groups = None
		self.param = None
	
	def check_required(self, settings, version):
		for field in ['category', 'score', 'order', 'update_mode', 'groups', 'param']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.category = stream.u32()
		self.score = stream.u32()
		self.order = stream.u8()
		self.update_mode = stream.u8()
		self.groups = stream.list(stream.u8)
		self.param = stream.u64()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.category)
		stream.u32(self.score)
		stream.u8(self.order)
		stream.u8(self.update_mode)
		stream.list(self.groups, stream.u8)
		stream.u64(self.param)


class RankingChangeAttributesParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.flags = None
		self.groups = None
		self.param = None
	
	def check_required(self, settings, version):
		for field in ['flags', 'groups', 'param']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.flags = stream.u8()
		self.groups = stream.list(stream.u8)
		self.param = stream.u64()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u8(self.flags)
		stream.list(self.groups, stream.u8)
		stream.u64(self.param)


class CommonDataList(common.Structure):
	def __init__(self):
		super().__init__()
		self.data = None
	
	def check_required(self, settings, version):
		for field in ['data']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.data = stream.list(stream.qbuffer)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.list(self.data, stream.qbuffer)


class CompetitionRankingGetScoreParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.id = None
		self.range = common.ResultRange()
	
	def check_required(self, settings, version):
		for field in ['id']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.id = stream.u32()
		self.range = stream.extract(common.ResultRange)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.id)
		stream.add(self.range)


class CompetitionRankingInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.id = None
		self.num_participants = None
		self.team_scores = None
	
	def check_required(self, settings, version):
		for field in ['id', 'num_participants', 'team_scores']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.id = stream.u32()
		self.num_participants = stream.u32()
		self.team_scores = stream.list(stream.u32)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.id)
		stream.u32(self.num_participants)
		stream.list(self.team_scores, stream.u32)


class CompetitionRankingInfoGetParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.rank_order = None
		self.range = common.ResultRange()
	
	def check_required(self, settings, version):
		for field in ['rank_order']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.rank_order = stream.u8()
		self.range = stream.extract(common.ResultRange)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u8(self.rank_order)
		stream.add(self.range)


class CompetitionRankingScoreData(common.Structure):
	def __init__(self):
		super().__init__()
		self.rank = None
		self.pid = None
		self.score = None
		self.last_update = None
		self.team_id = 255
		self.metadata = None
	
	def check_required(self, settings, version):
		for field in ['rank', 'pid', 'score', 'last_update', 'metadata']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.rank = stream.u32()
		self.pid = stream.pid()
		self.score = stream.u32()
		self.last_update = stream.datetime()
		self.team_id = stream.u8()
		self.metadata = stream.qbuffer()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.rank)
		stream.pid(self.pid)
		stream.u32(self.score)
		stream.datetime(self.last_update)
		stream.u8(self.team_id)
		stream.qbuffer(self.metadata)


class CompetitionRankingScoreInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.season_id = None
		self.scores = None
		self.num_participants = None
		self.team_scores = None
	
	def check_required(self, settings, version):
		for field in ['season_id', 'scores', 'num_participants', 'team_scores']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.season_id = stream.u32()
		self.scores = stream.list(CompetitionRankingScoreData)
		self.num_participants = stream.u32()
		self.team_scores = stream.list(stream.u32)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.season_id)
		stream.list(self.scores, stream.add)
		stream.u32(self.num_participants)
		stream.list(self.team_scores, stream.u32)


class CompetitionRankingUploadScoreParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.id = None
		self.season_id = None
		self.unk3 = None
		self.score = None
		self.team_id = None
		self.team_score = None
		self.is_first_upload = None
		self.metadata = None
	
	def check_required(self, settings, version):
		for field in ['id', 'season_id', 'unk3', 'score', 'team_id', 'team_score', 'is_first_upload', 'metadata']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.id = stream.u32()
		self.season_id = stream.u32()
		self.unk3 = stream.u32()
		self.score = stream.u32()
		self.team_id = stream.u8()
		self.team_score = stream.u32()
		self.is_first_upload = stream.bool()
		self.metadata = stream.qbuffer()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.id)
		stream.u32(self.season_id)
		stream.u32(self.unk3)
		stream.u32(self.score)
		stream.u8(self.team_id)
		stream.u32(self.team_score)
		stream.bool(self.is_first_upload)
		stream.qbuffer(self.metadata)


class ScorePack(common.Structure):
	def __init__(self):
		super().__init__()
		self.data = None
	
	def check_required(self, settings, version):
		for field in ['data']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.data = stream.list(stream.qbuffer)
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.list(self.data, stream.qbuffer)


class RankingProtocolMK8D:
	METHOD_UPLOAD_SCORE = 1
	METHOD_DELETE_SCORE = 2
	METHOD_DELETE_ALL_SCORES = 3
	METHOD_UPLOAD_COMMON_DATA = 4
	METHOD_DELETE_COMMON_DATA = 5
	METHOD_GET_COMMON_DATA = 6
	METHOD_CHANGE_ATTRIBUTES = 7
	METHOD_CHANGE_ALL_ATTRIBUTES = 8
	METHOD_GET_RANKING = 9
	METHOD_GET_APPROX_ORDER = 10
	METHOD_GET_STATS = 11
	METHOD_GET_RANKING_BY_PID_LIST = 12
	METHOD_GET_RANKING_BY_UNIQUE_ID_LIST = 13
	METHOD_GET_CACHED_TOPX_RANKING = 14
	METHOD_GET_CACHED_TOPX_RANKINGS = 15
	METHOD_GET_COMPETITION_RANKING_SCORE = 16
	METHOD_UPLOAD_COMPETITION_RANKING_SCORE = 17
	METHOD_GET_COMPETITION_INFO = 18
	METHOD_UPLOAD_SCORE_PACK = 19
	METHOD_GET_SCORE_PACK = 20
	METHOD_EXECUTE_DELETE_SCORE_JOB = 21
	METHOD_GET_COMMMON_DATA_BY_PID_LIST = 22
	
	PROTOCOL_ID = 0x70


class RankingClientMK8D(RankingProtocolMK8D):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def upload_score(self, score_data, unique_id):
		logger.info("RankingClientMK8D.upload_score()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(score_data)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPLOAD_SCORE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.upload_score -> done")
	
	async def delete_score(self, category, unique_id):
		logger.info("RankingClientMK8D.delete_score()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(category)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_SCORE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.delete_score -> done")
	
	async def delete_all_scores(self, unique_id):
		logger.info("RankingClientMK8D.delete_all_scores()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_ALL_SCORES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.delete_all_scores -> done")
	
	async def upload_common_data(self, common_data, unique_id):
		logger.info("RankingClientMK8D.upload_common_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.buffer(common_data)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPLOAD_COMMON_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.upload_common_data -> done")
	
	async def delete_common_data(self, unique_id):
		logger.info("RankingClientMK8D.delete_common_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_DELETE_COMMON_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.delete_common_data -> done")
	
	async def get_common_data(self, unique_id):
		logger.info("RankingClientMK8D.get_common_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_COMMON_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		data = stream.buffer()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_common_data -> done")
		return data
	
	async def change_attributes(self, category, param, unique_id):
		logger.info("RankingClientMK8D.change_attributes()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(category)
		stream.add(param)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CHANGE_ATTRIBUTES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.change_attributes -> done")
	
	async def change_all_attributes(self, param, unique_id):
		logger.info("RankingClientMK8D.change_all_attributes()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_CHANGE_ALL_ATTRIBUTES, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.change_all_attributes -> done")
	
	async def get_ranking(self, mode, category, order, unique_id, pid):
		logger.info("RankingClientMK8D.get_ranking()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u8(mode)
		stream.u32(category)
		stream.add(order)
		stream.u64(unique_id)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.extract(RankingResult)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_ranking -> done")
		return result
	
	async def get_approx_order(self, category, order, score, unique_id, pid):
		logger.info("RankingClientMK8D.get_approx_order()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(category)
		stream.add(order)
		stream.u32(score)
		stream.u64(unique_id)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_APPROX_ORDER, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		order = stream.u32()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_approx_order -> done")
		return order
	
	async def get_stats(self, category, order, flags):
		logger.info("RankingClientMK8D.get_stats()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(category)
		stream.add(order)
		stream.u32(flags)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_STATS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		stats = stream.extract(RankingStats)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_stats -> done")
		return stats
	
	async def get_ranking_by_pid_list(self, pids, mode, category, order, unique_id):
		logger.info("RankingClientMK8D.get_ranking_by_pid_list()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(pids, stream.pid)
		stream.u8(mode)
		stream.u32(category)
		stream.add(order)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING_BY_PID_LIST, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.extract(RankingResult)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_ranking_by_pid_list -> done")
		return result
	
	async def get_ranking_by_unique_id_list(self, ids, mode, category, order, unique_id):
		logger.info("RankingClientMK8D.get_ranking_by_unique_id_list()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(ids, stream.u64)
		stream.u8(mode)
		stream.u32(category)
		stream.add(order)
		stream.u64(unique_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_RANKING_BY_UNIQUE_ID_LIST, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.extract(RankingResult)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_ranking_by_unique_id_list -> done")
		return result
	
	async def get_cached_topx_ranking(self, category, order):
		logger.info("RankingClientMK8D.get_cached_topx_ranking()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(category)
		stream.add(order)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_CACHED_TOPX_RANKING, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.extract(RankingCachedResult)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_cached_topx_ranking -> done")
		return result
	
	async def get_cached_topx_rankings(self, categories, order):
		logger.info("RankingClientMK8D.get_cached_topx_rankings()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(categories, stream.u32)
		stream.list(order, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_CACHED_TOPX_RANKINGS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		results = stream.list(RankingCachedResult)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_cached_topx_rankings -> done")
		return results
	
	async def get_competition_ranking_score(self, param):
		logger.info("RankingClientMK8D.get_competition_ranking_score()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_COMPETITION_RANKING_SCORE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.list(CompetitionRankingScoreInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_competition_ranking_score -> done")
		return info
	
	async def upload_competition_ranking_score(self, param):
		logger.info("RankingClientMK8D.upload_competition_ranking_score()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPLOAD_COMPETITION_RANKING_SCORE, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.bool()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.upload_competition_ranking_score -> done")
		return result
	
	async def get_competition_info(self, param):
		logger.info("RankingClientMK8D.get_competition_info()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_COMPETITION_INFO, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.list(CompetitionRankingInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_competition_info -> done")
		return info
	
	async def upload_score_pack(self, score_data, metadata):
		logger.info("RankingClientMK8D.upload_score_pack()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(score_data)
		stream.qbuffer(metadata)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPLOAD_SCORE_PACK, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.upload_score_pack -> done")
	
	async def get_score_pack(self, pids, category):
		logger.info("RankingClientMK8D.get_score_pack()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(pids, stream.pid)
		stream.u32(category)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_SCORE_PACK, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.extract(ScorePack)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_score_pack -> done")
		return result
	
	async def get_commmon_data_by_pid_list(self, pids):
		logger.info("RankingClientMK8D.get_commmon_data_by_pid_list()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(pids, stream.pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_COMMMON_DATA_BY_PID_LIST, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		list = stream.extract(CommonDataList)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("RankingClientMK8D.get_commmon_data_by_pid_list -> done")
		return list


class RankingServerMK8D(RankingProtocolMK8D):
	def __init__(self):
		self.methods = {
			self.METHOD_UPLOAD_SCORE: self.handle_upload_score,
			self.METHOD_DELETE_SCORE: self.handle_delete_score,
			self.METHOD_DELETE_ALL_SCORES: self.handle_delete_all_scores,
			self.METHOD_UPLOAD_COMMON_DATA: self.handle_upload_common_data,
			self.METHOD_DELETE_COMMON_DATA: self.handle_delete_common_data,
			self.METHOD_GET_COMMON_DATA: self.handle_get_common_data,
			self.METHOD_CHANGE_ATTRIBUTES: self.handle_change_attributes,
			self.METHOD_CHANGE_ALL_ATTRIBUTES: self.handle_change_all_attributes,
			self.METHOD_GET_RANKING: self.handle_get_ranking,
			self.METHOD_GET_APPROX_ORDER: self.handle_get_approx_order,
			self.METHOD_GET_STATS: self.handle_get_stats,
			self.METHOD_GET_RANKING_BY_PID_LIST: self.handle_get_ranking_by_pid_list,
			self.METHOD_GET_RANKING_BY_UNIQUE_ID_LIST: self.handle_get_ranking_by_unique_id_list,
			self.METHOD_GET_CACHED_TOPX_RANKING: self.handle_get_cached_topx_ranking,
			self.METHOD_GET_CACHED_TOPX_RANKINGS: self.handle_get_cached_topx_rankings,
			self.METHOD_GET_COMPETITION_RANKING_SCORE: self.handle_get_competition_ranking_score,
			self.METHOD_UPLOAD_COMPETITION_RANKING_SCORE: self.handle_upload_competition_ranking_score,
			self.METHOD_GET_COMPETITION_INFO: self.handle_get_competition_info,
			self.METHOD_UPLOAD_SCORE_PACK: self.handle_upload_score_pack,
			self.METHOD_GET_SCORE_PACK: self.handle_get_score_pack,
			self.METHOD_EXECUTE_DELETE_SCORE_JOB: self.handle_execute_delete_score_job,
			self.METHOD_GET_COMMMON_DATA_BY_PID_LIST: self.handle_get_commmon_data_by_pid_list,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on RankingServerMK8D: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_upload_score(self, client, input, output):
		logger.info("RankingServerMK8D.upload_score()")
		#--- request ---
		score_data = input.extract(RankingScoreData)
		unique_id = input.u64()
		await self.upload_score(client, score_data, unique_id)
	
	async def handle_delete_score(self, client, input, output):
		logger.info("RankingServerMK8D.delete_score()")
		#--- request ---
		category = input.u32()
		unique_id = input.u64()
		await self.delete_score(client, category, unique_id)
	
	async def handle_delete_all_scores(self, client, input, output):
		logger.info("RankingServerMK8D.delete_all_scores()")
		#--- request ---
		unique_id = input.u64()
		await self.delete_all_scores(client, unique_id)
	
	async def handle_upload_common_data(self, client, input, output):
		logger.info("RankingServerMK8D.upload_common_data()")
		#--- request ---
		common_data = input.buffer()
		unique_id = input.u64()
		await self.upload_common_data(client, common_data, unique_id)
	
	async def handle_delete_common_data(self, client, input, output):
		logger.info("RankingServerMK8D.delete_common_data()")
		#--- request ---
		unique_id = input.u64()
		await self.delete_common_data(client, unique_id)
	
	async def handle_get_common_data(self, client, input, output):
		logger.info("RankingServerMK8D.get_common_data()")
		#--- request ---
		unique_id = input.u64()
		response = await self.get_common_data(client, unique_id)
		
		#--- response ---
		if not isinstance(response, bytes):
			raise RuntimeError("Expected bytes, got %s" %response.__class__.__name__)
		output.buffer(response)
	
	async def handle_change_attributes(self, client, input, output):
		logger.info("RankingServerMK8D.change_attributes()")
		#--- request ---
		category = input.u32()
		param = input.extract(RankingChangeAttributesParam)
		unique_id = input.u64()
		await self.change_attributes(client, category, param, unique_id)
	
	async def handle_change_all_attributes(self, client, input, output):
		logger.info("RankingServerMK8D.change_all_attributes()")
		#--- request ---
		param = input.extract(RankingChangeAttributesParam)
		unique_id = input.u64()
		await self.change_all_attributes(client, param, unique_id)
	
	async def handle_get_ranking(self, client, input, output):
		logger.info("RankingServerMK8D.get_ranking()")
		#--- request ---
		mode = input.u8()
		category = input.u32()
		order = input.extract(RankingOrderParam)
		unique_id = input.u64()
		pid = input.pid()
		response = await self.get_ranking(client, mode, category, order, unique_id, pid)
		
		#--- response ---
		if not isinstance(response, RankingResult):
			raise RuntimeError("Expected RankingResult, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_approx_order(self, client, input, output):
		logger.info("RankingServerMK8D.get_approx_order()")
		#--- request ---
		category = input.u32()
		order = input.extract(RankingOrderParam)
		score = input.u32()
		unique_id = input.u64()
		pid = input.pid()
		response = await self.get_approx_order(client, category, order, score, unique_id, pid)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.u32(response)
	
	async def handle_get_stats(self, client, input, output):
		logger.info("RankingServerMK8D.get_stats()")
		#--- request ---
		category = input.u32()
		order = input.extract(RankingOrderParam)
		flags = input.u32()
		response = await self.get_stats(client, category, order, flags)
		
		#--- response ---
		if not isinstance(response, RankingStats):
			raise RuntimeError("Expected RankingStats, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_ranking_by_pid_list(self, client, input, output):
		logger.info("RankingServerMK8D.get_ranking_by_pid_list()")
		#--- request ---
		pids = input.list(input.pid)
		mode = input.u8()
		category = input.u32()
		order = input.extract(RankingOrderParam)
		unique_id = input.u64()
		response = await self.get_ranking_by_pid_list(client, pids, mode, category, order, unique_id)
		
		#--- response ---
		if not isinstance(response, RankingResult):
			raise RuntimeError("Expected RankingResult, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_ranking_by_unique_id_list(self, client, input, output):
		logger.info("RankingServerMK8D.get_ranking_by_unique_id_list()")
		#--- request ---
		ids = input.list(input.u64)
		mode = input.u8()
		category = input.u32()
		order = input.extract(RankingOrderParam)
		unique_id = input.u64()
		response = await self.get_ranking_by_unique_id_list(client, ids, mode, category, order, unique_id)
		
		#--- response ---
		if not isinstance(response, RankingResult):
			raise RuntimeError("Expected RankingResult, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_cached_topx_ranking(self, client, input, output):
		logger.info("RankingServerMK8D.get_cached_topx_ranking()")
		#--- request ---
		category = input.u32()
		order = input.extract(RankingOrderParam)
		response = await self.get_cached_topx_ranking(client, category, order)
		
		#--- response ---
		if not isinstance(response, RankingCachedResult):
			raise RuntimeError("Expected RankingCachedResult, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_get_cached_topx_rankings(self, client, input, output):
		logger.info("RankingServerMK8D.get_cached_topx_rankings()")
		#--- request ---
		categories = input.list(input.u32)
		order = input.list(RankingOrderParam)
		response = await self.get_cached_topx_rankings(client, categories, order)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_get_competition_ranking_score(self, client, input, output):
		logger.info("RankingServerMK8D.get_competition_ranking_score()")
		#--- request ---
		param = input.extract(CompetitionRankingGetScoreParam)
		response = await self.get_competition_ranking_score(client, param)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_upload_competition_ranking_score(self, client, input, output):
		logger.info("RankingServerMK8D.upload_competition_ranking_score()")
		#--- request ---
		param = input.extract(CompetitionRankingUploadScoreParam)
		response = await self.upload_competition_ranking_score(client, param)
		
		#--- response ---
		if not isinstance(response, bool):
			raise RuntimeError("Expected bool, got %s" %response.__class__.__name__)
		output.bool(response)
	
	async def handle_get_competition_info(self, client, input, output):
		logger.info("RankingServerMK8D.get_competition_info()")
		#--- request ---
		param = input.extract(CompetitionRankingInfoGetParam)
		response = await self.get_competition_info(client, param)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def handle_upload_score_pack(self, client, input, output):
		logger.info("RankingServerMK8D.upload_score_pack()")
		#--- request ---
		score_data = input.extract(RankingScoreData)
		metadata = input.qbuffer()
		await self.upload_score_pack(client, score_data, metadata)
	
	async def handle_get_score_pack(self, client, input, output):
		logger.info("RankingServerMK8D.get_score_pack()")
		#--- request ---
		pids = input.list(input.pid)
		category = input.u32()
		response = await self.get_score_pack(client, pids, category)
		
		#--- response ---
		if not isinstance(response, ScorePack):
			raise RuntimeError("Expected ScorePack, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def handle_execute_delete_score_job(self, client, input, output):
		logger.warning("RankingServerMK8D.execute_delete_score_job is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_get_commmon_data_by_pid_list(self, client, input, output):
		logger.info("RankingServerMK8D.get_commmon_data_by_pid_list()")
		#--- request ---
		pids = input.list(input.pid)
		response = await self.get_commmon_data_by_pid_list(client, pids)
		
		#--- response ---
		if not isinstance(response, CommonDataList):
			raise RuntimeError("Expected CommonDataList, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def upload_score(self, *args):
		logger.warning("RankingServerMK8D.upload_score not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_score(self, *args):
		logger.warning("RankingServerMK8D.delete_score not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_all_scores(self, *args):
		logger.warning("RankingServerMK8D.delete_all_scores not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def upload_common_data(self, *args):
		logger.warning("RankingServerMK8D.upload_common_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def delete_common_data(self, *args):
		logger.warning("RankingServerMK8D.delete_common_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_common_data(self, *args):
		logger.warning("RankingServerMK8D.get_common_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def change_attributes(self, *args):
		logger.warning("RankingServerMK8D.change_attributes not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def change_all_attributes(self, *args):
		logger.warning("RankingServerMK8D.change_all_attributes not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking(self, *args):
		logger.warning("RankingServerMK8D.get_ranking not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_approx_order(self, *args):
		logger.warning("RankingServerMK8D.get_approx_order not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_stats(self, *args):
		logger.warning("RankingServerMK8D.get_stats not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking_by_pid_list(self, *args):
		logger.warning("RankingServerMK8D.get_ranking_by_pid_list not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_ranking_by_unique_id_list(self, *args):
		logger.warning("RankingServerMK8D.get_ranking_by_unique_id_list not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_cached_topx_ranking(self, *args):
		logger.warning("RankingServerMK8D.get_cached_topx_ranking not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_cached_topx_rankings(self, *args):
		logger.warning("RankingServerMK8D.get_cached_topx_rankings not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_competition_ranking_score(self, *args):
		logger.warning("RankingServerMK8D.get_competition_ranking_score not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def upload_competition_ranking_score(self, *args):
		logger.warning("RankingServerMK8D.upload_competition_ranking_score not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_competition_info(self, *args):
		logger.warning("RankingServerMK8D.get_competition_info not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def upload_score_pack(self, *args):
		logger.warning("RankingServerMK8D.upload_score_pack not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_score_pack(self, *args):
		logger.warning("RankingServerMK8D.get_score_pack not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_commmon_data_by_pid_list(self, *args):
		logger.warning("RankingServerMK8D.get_commmon_data_by_pid_list not implemented")
		raise common.RMCError("Core::NotImplemented")

